import typing

import attr

from .match import ImageMatchSettings
from .server import SessionType
from .utils.compat import basestring
from .utils.json_utils import JsonInclude

if typing.TYPE_CHECKING:
    from applitools.common import (
        DeviceName,
        IosDeviceName,
        BatchInfo,
        RenderingInfo,
        RectangleSize,
    )
    from typing import Optional, Text, Union

__all__ = ("RunningSession", "AppEnvironment", "SessionStartInfo")


@attr.s
class RunningSession(object):
    """
    Encapsulates data for the session currently running in the agent.
    """

    id = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    session_id = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    batch_id = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    baseline_id = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    url = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    rendering_info = attr.ib(default=None)  # type: Optional[RenderingInfo]
    is_new_session = attr.ib(
        default=False, metadata={JsonInclude.NAME: "isNew"}
    )  # type: bool


def _to_device_info(v):
    # type: (Optional[Union[DeviceName, IosDeviceName, Text]]) -> Text
    # import here because in global scope causing circular import reference
    from .ultrafastgrid.config import IosDeviceName, DeviceName  # noqa

    if isinstance(v, basestring):
        return v
    elif isinstance(v, DeviceName):
        return "{} (Chrome emulation)".format(v.value)
    elif isinstance(v, IosDeviceName):
        return v.value
    return "Desktop"


@attr.s
class AppEnvironment(object):
    os = attr.ib(
        default=None, metadata={JsonInclude.NON_NONE: True}
    )  # type:Optional[Text]
    hosting_app = attr.ib(
        default=None, metadata={JsonInclude.NON_NONE: True}
    )  # type:Optional[Text]
    display_size = attr.ib(
        default=None, metadata={JsonInclude.NON_NONE: True}
    )  # type:Optional[RectangleSize]
    device_info = attr.ib(
        default=None, converter=_to_device_info, metadata={JsonInclude.NON_NONE: True}
    )  # type:Optional[Text]
    os_info = attr.ib(
        default=None, metadata={JsonInclude.NON_NONE: True}
    )  # type:Optional[Text]
    hosting_app_info = attr.ib(
        default=None, metadata={JsonInclude.NON_NONE: True}
    )  # type:Optional[Text]
    inferred = attr.ib(
        default=None, metadata={JsonInclude.NON_NONE: True}
    )  # type:Optional[Text]

    @classmethod
    def from_inferred(cls, inferred):
        return cls(inferred=inferred)


@attr.s
class SessionStartInfo(object):
    JSON_NAME = "startInfo"
    agent_id = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    session_type = attr.ib(
        type=SessionType, metadata={JsonInclude.THIS: True}
    )  # type: SessionType
    app_id_or_name = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    ver_id = attr.ib(metadata={JsonInclude.THIS: True})  # type: Optional[Text]
    scenario_id_or_name = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    batch_info = attr.ib(metadata={JsonInclude.THIS: True})  # type: BatchInfo
    baseline_env_name = attr.ib(metadata={JsonInclude.NON_NONE: True})  # type: Text
    environment_name = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    environment = attr.ib(
        type=AppEnvironment, metadata={JsonInclude.THIS: True}
    )  # type: AppEnvironment
    default_match_settings = attr.ib(
        type=ImageMatchSettings, metadata={JsonInclude.THIS: True}
    )  # type: ImageMatchSettings
    branch_name = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    parent_branch_name = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    baseline_branch_name = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    save_diffs = attr.ib(metadata={JsonInclude.THIS: True})  # type: bool
    render = attr.ib(metadata={JsonInclude.THIS: True})  # type: bool
    properties = attr.ib(metadata={JsonInclude.THIS: True})  # type: list



import base64
import typing
from enum import Enum

import attr

from applitools.common import logger
from applitools.common.geometry import RectangleSize, Region
from applitools.common.selenium.misc import BrowserType
from applitools.common.utils import general_utils, json_utils
from applitools.common.utils.json_utils import JsonInclude
from .render_browser_info import (
    IosDeviceInfo,
    IRenderBrowserInfo,
    ChromeEmulationInfo,
    EmulationBaseInfo,
)

if typing.TYPE_CHECKING:
    from typing import List, Text, Dict, Optional, Callable, Union, Any
    from requests import Response
    from applitools.common.utils.custom_types import Num
    from applitools.selenium.visual_grid.vg_task import VGTask

__all__ = (
    "RenderStatus",
    "RenderingInfo",
    "VisualGridSelector",
    "RenderInfo",
    "RGridDom",
    "RunningRender",
    "VGResource",
    "RenderRequest",
    "RenderStatusResults",
)


class RenderStatus(Enum):
    NEED_MORE_RESOURCE = "need-more-resources"
    RENDERING = "rendering"
    RENDERED = "rendered"
    ERROR = "error"
    NEED_MORE_DOM = "need-more-dom"
    INTERNAL_FAILURE = "internal failure"


@attr.s
class VGRegion(object):
    x = attr.ib(default=None, metadata={JsonInclude.THIS: True})
    y = attr.ib(default=None, metadata={JsonInclude.THIS: True})
    width = attr.ib(default=None, metadata={JsonInclude.THIS: True})
    height = attr.ib(default=None, metadata={JsonInclude.THIS: True})
    error = attr.ib(default=None, metadata={JsonInclude.THIS: True})

    def to_region(self):
        return Region(self.x, self.y, self.width, self.height)

    def offset(self, dx, dy):
        return self.to_region().offset(dx, dy)


@attr.s(frozen=True)
class RenderingInfo(object):
    service_url = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: Optional[Text]
    access_token = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: Optional[Text]
    results_url = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: Optional[Text]
    stitching_service_url = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: Optional[Text]


@attr.s(hash=True, slots=True)
class VisualGridSelector(object):
    selector = attr.ib(metadata={JsonInclude.NON_NONE: True})  # type: Text
    category = attr.ib()  # type: Union[Text, Any]
    type = attr.ib(default="xpath", metadata={JsonInclude.THIS: True})


@attr.s
class RenderInfo(object):
    width = attr.ib(metadata={JsonInclude.NON_NONE: True})  # type: Optional[Num]
    height = attr.ib(metadata={JsonInclude.NON_NONE: True})  # type: Optional[Num]
    size_mode = attr.ib(metadata={JsonInclude.NON_NONE: True})  # type: Optional[Text]
    region = attr.ib(
        default=None, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[Region]
    selector = attr.ib(
        default=None, type=VisualGridSelector, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[VisualGridSelector]
    emulation_info = attr.ib(
        default=None, type=EmulationBaseInfo, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[EmulationBaseInfo]
    ios_device_info = attr.ib(
        default=None, type=IosDeviceInfo, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[IosDeviceInfo]

    @classmethod
    def from_(
        cls,
        size_mode,  # type: Optional[Text]
        region,  # type: Optional[Region]
        selector,  # type: Optional[VisualGridSelector]
        render_browser_info,  # type: IRenderBrowserInfo
    ):
        # type: (...) -> RenderInfo
        ios_device_info = None
        emulation_info = None
        if isinstance(render_browser_info, IosDeviceInfo):
            ios_device_info = render_browser_info
        if isinstance(render_browser_info, ChromeEmulationInfo):
            emulation_info = render_browser_info
        return cls(
            width=render_browser_info.width,
            height=render_browser_info.height,
            size_mode=size_mode,
            region=region,
            selector=selector,
            emulation_info=emulation_info,
            ios_device_info=ios_device_info,
        )


@attr.s
class RGridDom(object):
    CONTENT_TYPE = "x-applitools-html/cdt"  # type: Text

    dom_nodes = attr.ib(repr=False)  # type: List[dict]
    resources = attr.ib()  # type: Dict[Text, VGResource]
    url = attr.ib()  # type: Text
    msg = attr.ib(default=None)  # type: Text
    hash = attr.ib(init=False, metadata={JsonInclude.THIS: True})  # type: Text
    hash_format = attr.ib(
        default="sha256", metadata={JsonInclude.THIS: True}
    )  # type: Text

    def __attrs_post_init__(self):
        # TODO: add proper hash
        self.hash = general_utils.get_sha256_hash(self.content.encode("utf-8"))

    @property
    def resource(self):
        # type: () -> VGResource
        return VGResource(
            self.url,
            self.CONTENT_TYPE,
            self.content.encode("utf-8"),
            "RGridDom {}".format(self.msg),
        )

    @property
    def content(self):
        # type: () -> Text
        data = {"resources": self.resources, "domNodes": self.dom_nodes}
        return json_utils.to_json(data)


@attr.s(slots=True)
class VGResource(object):
    url = attr.ib()  # type: Text
    content_type = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    content = attr.ib(repr=False)  # type: bytes
    msg = attr.ib(default=None)  # type: Optional[Text]
    hash = attr.ib(init=False, metadata={JsonInclude.THIS: True})  # type: Text
    hash_format = attr.ib(
        init=False, default="sha256", metadata={JsonInclude.THIS: True}
    )  # type: Text
    _handle_func = attr.ib(default=None)

    def __hash__(self):
        return self.hash

    def __attrs_post_init__(self):
        self.hash = general_utils.get_sha256_hash(self.content)
        if callable(self._handle_func):
            try:
                self._handle_func()
            except Exception:
                logger.exception(
                    "Exception has been appeared during processing"
                    " of VGResource({})".format(self.url),
                )

    @classmethod
    def EMPTY(cls, url):
        return cls(url, "application/empty-response", b"")

    @classmethod
    def from_blob(cls, blob, on_created=None):
        # type: (Dict, Callable) -> VGResource
        content = base64.b64decode(blob.get("value", ""))
        content_type = blob.get("type")
        url = blob.get("url")
        return cls(
            url,
            content_type,
            content,
            handle_func=lambda: on_created(content_type, content, url),
        )

    @classmethod
    def from_response(cls, url, response, on_created=None):
        # type: (Text, Response, Callable) -> VGResource
        if not response.ok:
            logger.debug(
                "We've got response code {} {} for URL {}".format(
                    response.status_code, response.reason, url
                )
            )
            return VGResource.EMPTY(url)

        content_type = response.headers.get("Content-Type")
        content = response.content
        return cls(
            url,
            content_type,
            content,
            handle_func=lambda: on_created(content_type, content, url),
        )


@attr.s
class RenderRequest(object):
    webhook = attr.ib(metadata={JsonInclude.NON_NONE: True})  # type: Text
    agent_id = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    url = attr.ib(metadata={JsonInclude.NON_NONE: True})  # type: Text
    stitching_service = attr.ib(metadata={JsonInclude.NON_NONE: True})  # type: Text
    dom = attr.ib(repr=False, metadata={JsonInclude.NON_NONE: True})  # type: RGridDom
    resources = attr.ib(repr=False, metadata={JsonInclude.NON_NONE: True})  # type: dict
    render_info = attr.ib(metadata={JsonInclude.THIS: True})  # type: RenderInfo
    platform_name = attr.ib()  # type: Text
    browser_name = attr.ib()  # type: Text
    script_hooks = attr.ib(
        default=dict, metadata={JsonInclude.NON_NONE: True}
    )  # type: Dict
    selectors_to_find_regions_for = attr.ib(
        factory=list, metadata={JsonInclude.THIS: True}
    )  # type: List[VisualGridSelector]
    send_dom = attr.ib(
        default=False, metadata={JsonInclude.NON_NONE: True}
    )  # type: bool
    render_id = attr.ib(
        default=None, repr=True, metadata={JsonInclude.THIS: True}
    )  # type: Optional[Text]
    task = attr.ib(default=None)  # type: Optional[VGTask]
    browser = attr.ib(init=False, default=None, metadata={JsonInclude.NON_NONE: True})
    platform = attr.ib(
        init=False, default=None, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[Dict]

    def __attrs_post_init__(self):
        if self.browser_name is None:
            return
        self.browser = {"name": self.browser_name}
        self.platform = {"name": self.platform_name}


@attr.s(hash=True)
class RunningRender(object):
    render_id = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: Optional[Text]
    job_id = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: Optional[Text]
    render_status = attr.ib(
        default=None,
        converter=attr.converters.optional(RenderStatus),
        metadata={JsonInclude.THIS: True},
    )  # type: RenderStatus
    need_more_resources = attr.ib(
        default=None, hash=False, metadata={JsonInclude.THIS: True}
    )  # type: List[Text]
    need_more_dom = attr.ib(
        default=None, hash=False, metadata={JsonInclude.THIS: True}
    )  # type: bool


@attr.s
class RenderStatusResults(object):
    status = attr.ib(
        default=None,
        type=RenderStatus,
        converter=attr.converters.optional(RenderStatus),
        metadata={JsonInclude.THIS: True},
    )  # type: Optional[RenderStatus]
    dom_location = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: Optional[Text]
    user_agent = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: Optional[Text]
    image_location = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: Optional[Text]
    os = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: Optional[Text]
    error = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: Optional[Text]
    selector_regions = attr.ib(
        factory=list, type=VGRegion, metadata={JsonInclude.THIS: True}
    )  # type: List[VGRegion]
    device_size = attr.ib(
        default=None,
        type=RectangleSize,
        converter=attr.converters.optional(RectangleSize.from_),
        metadata={JsonInclude.THIS: True},
    )  # type: Optional[RectangleSize]
    retry_count = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: Optional[int]
    render_id = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: Optional[Text]
    visual_viewport = attr.ib(
        default=None, type=RectangleSize, metadata={JsonInclude.THIS: True}
    )  # type: Optional[RectangleSize]



"""
Compatibility layer between Python 2 and 3
"""
from __future__ import absolute_import

import abc
import inspect
import io
import sys
from gzip import GzipFile

__all__ = (
    "ABC",
    "range",
    "iteritems",
    "urlparse",
    "urljoin",
    "quote_plus",
    "Queue",
    "gzip_compress",
    "range",
    "basestring",
    "urlencode",
    "parse_qs",
    "urlsplit",
    "urlunsplit",
    "raise_from",
)


def _get_caller_globals_and_locals():
    """
    Returns the globals and locals of the calling frame.
    """
    caller_frame = inspect.stack()[2]
    myglobals = caller_frame[0].f_globals
    mylocals = caller_frame[0].f_locals
    return myglobals, mylocals


PY3 = sys.version_info[:1] >= (3,)

if PY3:
    from urllib.parse import (
        urlparse,
        urljoin,
        urlencode,
        parse_qs,
        urlsplit,
        urlunsplit,
        quote_plus,
    )  # noqa
    from gzip import compress as gzip_compress  # noqa
    from queue import Queue  # noqa

    basestring = str
    ABC = abc.ABC
    range = range  # type: ignore

    def raise_from(exc, cause):
        """
        Equivalent to:

            raise EXCEPTION from CAUSE

        on Python 3. (See PEP 3134).
        """
        myglobals, mylocals = _get_caller_globals_and_locals()

        # We pass the exception and cause along with other globals
        # when we exec():
        myglobals = myglobals.copy()
        myglobals["__python_future_raise_from_exc"] = exc
        myglobals["__python_future_raise_from_cause"] = cause
        execstr = (
            "raise __python_future_raise_from_exc from __python_future_raise_from_cause"
        )

        exec(execstr, myglobals, mylocals)


else:
    from urlparse import (
        urlparse,
        urljoin,
        parse_qs,
        urlsplit,
        urlunsplit,
    )  # noqa
    from urllib import urlencode, quote_plus  # noqa
    from Queue import Queue  # noqa

    basestring = basestring
    ABC = abc.ABCMeta(str("ABC"), (), {})
    range = xrange  # type: ignore  # noqa

    def gzip_compress(data, compresslevel=9):
        """Compress data in one shot and return the compressed string.
        Optional argument is the compression level, in range of 0-9.
        """
        buf = io.BytesIO()
        with GzipFile(fileobj=buf, mode="wb", compresslevel=compresslevel) as f:
            f.write(data)
        return buf.getvalue()

    def raise_from(exc, cause):
        """
        Equivalent to:

            raise EXCEPTION from CAUSE

        on Python 3. (See PEP 3134).
        """
        # Is either arg an exception class (e.g. IndexError) rather than
        # instance (e.g. IndexError('my message here')? If so, pass the
        # name of the class undisturbed through to "raise ... from ...".
        if isinstance(exc, type) and issubclass(exc, Exception):
            e = exc()
            # exc = exc.__name__
            # execstr = "e = " + _repr_strip(exc) + "()"
            # myglobals, mylocals = _get_caller_globals_and_locals()
            # exec(execstr, myglobals, mylocals)
        else:
            e = exc
        e.__suppress_context__ = False
        if isinstance(cause, type) and issubclass(cause, Exception):
            e.__cause__ = cause()
            e.__suppress_context__ = True
        elif cause is None:
            e.__cause__ = None
            e.__suppress_context__ = True
        elif isinstance(cause, BaseException):
            e.__cause__ = cause
            e.__suppress_context__ = True
        else:
            raise TypeError("exception causes must derive from BaseException")
        e.__context__ = sys.exc_info()[1]
        raise e


def iteritems(dct):
    return (getattr(dct, "iteritems", None) or dct.items)()



from . import argument_guard, compat, datetime_utils, image_utils  # noqa
from .compat import (  # type: ignore # noqa
    ABC,
    gzip_compress,
    iteritems,
    parse_qs,
    range,
    urlencode,
    urljoin,
    urlparse,
    urlsplit,
    urlunsplit,
    quote_plus,
)
from .datetime_utils import (  # type: ignore # noqa
    UTC,
    current_time_in_rfc1123,
    to_rfc1123_datetime,
)
from .general_utils import cached_property, counted  # noqa
from .url_utils import apply_base_url, is_url_with_scheme, is_absolute_url  # noqa


__all__ = (
    compat.__all__  # noqa
    + url_utils.__all__  # noqa
    + ("image_utils", "argument_guard", "datetime_utils", "counted")
)



from . import (  # noqa
    app_output,
    capture,
    config,
    errors,
    geometry,
    logger,
    match,
    match_window_data,
    metadata,
    selenium,
    server,
    test_results,
    ultrafastgrid,
)
from .app_output import AppOutput  # noqa
from .capture import EyesScreenshot  # noqa
from .config import BatchInfo, Configuration  # noqa
from .errors import (  # noqa
    CoordinatesTypeConversionError,
    DiffsFoundError,
    EyesError,
    NewTestError,
    OutOfBoundsError,
    TestFailedError,
)
from .geometry import (  # noqa
    CoordinatesType,
    Point,
    RectangleSize,
    Region,
    SubregionForStitching,
)
from .logger import FileLogger, NullLogger, StdoutLogger  # noqa
from .match import (  # noqa
    ExactMatchSettings,
    FloatingBounds,
    FloatingMatchSettings,
    ImageMatchSettings,
    MatchLevel,
    MatchResult,
)
from .match_window_data import MatchWindowData, Options  # noqa
from .metadata import AppEnvironment, RunningSession, SessionStartInfo  # noqa
from .scale_provider import ScaleProvider
from .selenium import StitchMode  # noqa
from .server import FailureReports, SessionType  # noqa
from .test_results import TestResultContainer, TestResults, TestResultsSummary  # noqa
from .ultrafastgrid.config import (  # noqa
    DeviceName,
    ScreenOrientation,
    IosDeviceName,
    ScreenOrientation,
)
from .ultrafastgrid.render_browser_info import (
    IosDeviceInfo,
    ChromeEmulationInfo,
    DesktopBrowserInfo,
    RenderBrowserInfo,
)
from .ultrafastgrid.render_request import (  # noqa
    RenderInfo,
    RenderingInfo,
    RenderRequest,
    RenderStatus,
    RenderStatusResults,
    RGridDom,
    RunningRender,
    VGResource,
    VisualGridSelector,
)
from .accessibility import (  # noqa
    AccessibilityRegionType,
    AccessibilitySettings,
    AccessibilityLevel,
    AccessibilityGuidelinesVersion,
)

__all__ = (
    logger.__all__  # noqa
    + errors.__all__  # noqa
    + geometry.__all__  # noqa
    + match.__all__  # noqa
    + metadata.__all__  # noqa
    + app_output.__all__  # noqa
    + capture.__all__  # noqa
    + match_window_data.__all__  # noqa
    + test_results.__all__  # noqa
    + server.__all__  # noqa
    + (
        "logger",
        "StitchMode",
        "ScaleProvider",
        "ChromeEmulationInfo",
        "DesktopBrowserInfo",
        "DeviceName",
        "ScreenOrientation",
        "RenderInfo",
        "RenderingInfo",
        "RenderRequest",
        "RenderStatus",
        "RenderStatusResults",
        "RGridDom",
        "RunningRender",
        "VGResource",
        "VisualGridSelector",
        "AccessibilityRegionType",
        "AccessibilitySettings",
        "AccessibilityLevel",
        "AccessibilityGuidelinesVersion",
    )  # noqa
)
